#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctime>
#include <sys/time.h>

#include "mpi_pdNCG.h"
#include "mpi_matvec.h"

#ifndef min
#define min(a, b) ((a) < (b) ? (a) : (b))
#endif

void PdNCG::pdNCG(int p, int size_p, long long n, long long m, long long k,
                  long long n_p, long long m_p, long long k_p, double *xopt_p, double *s_val_p,
                  double *b_noisy_p, double G[][2], double *x, double *y,
                  int maxpdNCGitr, int maxpcgitr, double mu, double tolpcg,
                  double tol)
{
    /* Allocate memory for primal direction */
    double *dx = new double[n_p];
    
    /* Other allocations */
    double *hessian_pHuber = new double[n_p];
    double *Dx             = new double[n_p];
    double *diag_AtA       = new double[n_p];
    
    double *re_local_and_norm_xopt_p  = new double[2];
    double *re_global_and_norm_xopt_p = new double[2];
    
    double alpha = 0.9;
    double re, re_local, y_temp;
    double norm_xopt_p, norm_xopt;
    double wall0, wall1;
    
    long long index;
    
    if (p == 0)
        wall0 = get_wall_time();
    
    /* Calculate diagonal of A^T*A */
    for (long long i = 0; i <n_p/2; i++)
    {
        index = 2 * i;
        
        for (long long j = 0; j < 2; j++)
        {
            diag_AtA[index + j] = 0;
            
            for (long long q = 0; q < 2; q++)
                diag_AtA[index + j] += pow(G[j][q], 2)*pow(s_val_p[index + q], 2);
        }
    }
    
//    for (long long j = 0; j < n_p; j++)
//        printf("diag_AtA_%d[%lld] = %e \n", p, j, diag_AtA[j]);
    
    for (int i = 0; i < maxpdNCGitr; i++)
    {
        /* Calculate gradient (Dx) for pseudo-Huber */
        for (long long j = 0; j < n_p; j++)
            Dx[j]   = x[j]/sqrt(pow(mu, 2) + pow(x[j], 2));
        
        /* Calculate Hessian for pseudo-Huber */
        for (long long j = 0; j <n_p; j++)
            hessian_pHuber[j] = (1 - Dx[j]*y[j])/sqrt(pow(mu, 2) + pow(x[j], 2));
        
//        for (long int j = 0; j < n_p; j++)
//            printf("hessian_pHuber_%d[%lld] = %e \n", p, j, hessian_pHuber[j]);
        
        /* Calculate approximate Newton direction for primal variables */
        direction(p, n_p, m_p, s_val_p, b_noisy_p, G, x, y, Dx, diag_AtA,
                  hessian_pHuber, dx, maxpcgitr, tolpcg);
        
        /* Update primal variables */
        for (long long j = 0; j < n_p; j++)
            x[j] = x[j] + alpha*dx[j];
        
        /* Calculate direction for dual variables */
        for (long long j = 0; j < n_p; j++)
        {
            y_temp    = y[j] + hessian_pHuber[j]*dx[j] - (y[j] - Dx[j]);
            y[j]      = sgn(y_temp)*min(fabs(y_temp), 1);
        }
        
        re_local    = 0;
        norm_xopt_p = 0;
        for (long long j = 0; j < n_p; j++)
        {
            re_local    += pow(x[j] - xopt_p[j], 2);
            norm_xopt_p += pow(xopt_p[j], 2);
        }
        
        re_local_and_norm_xopt_p[0] = re_local;
        re_local_and_norm_xopt_p[1] = norm_xopt_p;
        
        MPI_Allreduce(re_local_and_norm_xopt_p, re_global_and_norm_xopt_p, 2, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
        
        re = sqrt(re_global_and_norm_xopt_p[0]);
        norm_xopt = sqrt(re_global_and_norm_xopt_p[1]);
        
        re = re/norm_xopt;
        
        if (p == 0)
        {
            wall1 = get_wall_time();
            printf("re = %e at iteration %d at time %e \n", i, re, wall1 - wall0);
        }
        
        if(re <= tol)
        	break;
    }
    
    /* Allocate memory for primal direction */
    delete[] dx;
    
    /* Other allocations */
    delete[] hessian_pHuber;
    delete[] Dx;
    delete[] diag_AtA;
    
    delete[] re_local_and_norm_xopt_p;
    delete[] re_global_and_norm_xopt_p;
}

void PdNCG::direction(int p, long long n_p, long long m_p, double *s_val_p,
                      double *b_noisy_p, double G[][2], double *x, double *y,
                      double *Dx, double *diag_AtA,
                      double *hessian_pHuber, double *dx, int maxpcgitr, double tolpcg)
{
    Matvec matvec;
    
    /* Allocate memory for gradient of smoothed objective function*/
    //double *grad = new double[n_p];
    double *grad = new double[m_p];
    
    //double *Ax_b = new double[m_p];
    
    /* Calculate gradient of least-squares term */
    //matvec.matvec_A_forward(n_p, m_p, G, s_val_p, x, Ax_b);
    matvec.matvec_A_forward(n_p, m_p, G, s_val_p, x, grad);
    
    for (long long i = 0; i < m_p; i++)
        grad[i] = grad[i] - b_noisy_p[i];
        //Ax_b[i] = Ax_b[i] - b_noisy_p[i];
    
    //matvec.matvec_A_transpose(n_p, m_p, G, s_val_p, Ax_b, grad);
    matvec.matvec_A_transpose(n_p, m_p, G, s_val_p, grad, grad);
    
    for (long long i = 0; i < n_p; i++)
        grad[i] = Dx[i] + grad[i];
    
//    for (long long j = 0; j < n_p; j++)
//        printf("grad_%d[%lld] = %e \n", p, j, grad[j]);
    
    pcg(p, n_p, m_p, maxpcgitr, tolpcg, y, s_val_p, G, grad, diag_AtA, hessian_pHuber, dx);
    
//    for (long long j = 0; j < n_p; j++)
//        printf("dx_%d[%lld] = %e \n", p, j, dx[j]);
    
    delete[] grad;
    //delete[] Ax_b;

}

void PdNCG::pcg(int p, long long n_p, long long m_p, int maxpcgitr, double tolpcg, double *y, double *s_val_p,
                double G[][2], double *grad, double *diag_AtA, double *hessian_pHuber,
                double *d)
{
    double bnorm2;
    double rnorm2;
    double rz, rzold;
    double alpha, beta;
    double rz_local, rnorm2_local, bnorm2_local;
    double sz_local, sz;
    
    double *s = new double[n_p];
    double *r = new double[n_p];
    double *z = new double[m_p];
    
    double *local_bnorm2_and_rz_and_rnorm2  = new double[3];
    double *global_bnorm2_and_rz_and_rnorm2 = new double[3];
    double *local_rz_and_rnorm2             = new double[2];
    double *global_rz_and_rnorm2            = new double[2];
    
    for (long long i = 0; i < n_p; i++)
    {
        r[i] = -grad[i];
        d[i] = 0;
    }
    
    bnorm2_local = 0;
    for (long long i = 0; i < n_p; i++)
        bnorm2_local += pow(grad[i], 2);
    
    local_bnorm2_and_rz_and_rnorm2[0] = bnorm2_local;
    
    for (long long i = 0; i < n_p; i++)
        z[i] = r[i]/(diag_AtA[i] + hessian_pHuber[i]);
    
    for (long long i = 0; i < n_p; i++)
        s[i] = z[i];
    
    rz_local = 0;
    for (long long i = 0; i < n_p; i++)
        rz_local += z[i]*r[i];
    
    local_bnorm2_and_rz_and_rnorm2[1] = rz_local;
    
    rnorm2_local = 0;
    for (long long i = 0; i < n_p; i++)
        rnorm2_local += pow(r[i], 2);
    
    local_bnorm2_and_rz_and_rnorm2[2] = rnorm2_local;
    
    MPI_Allreduce(local_bnorm2_and_rz_and_rnorm2, global_bnorm2_and_rz_and_rnorm2, 3, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    
    bnorm2 = global_bnorm2_and_rz_and_rnorm2[0];
    rz     = global_bnorm2_and_rz_and_rnorm2[1];
    rnorm2 = global_bnorm2_and_rz_and_rnorm2[2];
    
    for (long long j = 0; j < maxpcgitr ; j++)
    {
        PdNCG::hesvec(n_p, m_p, s_val_p, G, hessian_pHuber, s, z);
        
        sz_local = 0;
        for (long long i = 0; i < n_p; i++)
            sz_local += s[i]*z[i];
        
        MPI_Allreduce(&sz_local, &sz, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
        
        alpha = rz / sz;
        
        for (long long i = 0; i < n_p; i++)
        {
            d[i] = alpha*s[i] + d[i];
            r[i] = r[i] - alpha*z[i];
        }
        
        for (long long i = 0; i < n_p; i++)
            z[i] = r[i]/(diag_AtA[i] + hessian_pHuber[i]);
        
        rzold = rz;
        
        rz_local = 0;
        for (long long i = 0; i < n_p; i++)
            rz_local += z[i]*r[i];
        
        local_rz_and_rnorm2[0] = rz_local;
        
        rnorm2_local = 0;
        for (long long i = 0; i < n_p; i++)
            rnorm2_local += pow(r[i], 2);
        
        local_rz_and_rnorm2[1] = rnorm2_local;
        
        MPI_Allreduce(local_rz_and_rnorm2, global_rz_and_rnorm2, 2, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
        
        rz     = global_rz_and_rnorm2[0];
        rnorm2 = global_rz_and_rnorm2[1];
        
        beta = -rz / rzold;
        
        for (long long i = 0; i < n_p; i++)
            s[i] = z[i] - beta*s[i];
        
//        if (p == 0)
//        {
//            printf("pcg iteration %d \n", j);
//            printf("pcg re %e \n", sqrt(rnorm2/bnorm2));
//        }
        
        if(rnorm2 <= bnorm2 * pow(tolpcg, 2))
        {
            if (p == 0)
            {
                printf("pcg iteration %d \n", j);
                printf("pcg re %e \n", sqrt(rnorm2/bnorm2));
            }
        	break;
        }
    }
    
    delete[] s;
    delete[] r;
    delete[] z;
    
    delete[] local_bnorm2_and_rz_and_rnorm2;
    delete[] global_bnorm2_and_rz_and_rnorm2;
    delete[] local_rz_and_rnorm2;
    delete[] global_rz_and_rnorm2;
}

void PdNCG::hesvec(long long n_p, long long m_p, double *s_val_p, double G[][2], double *hessian_pHuber,
                   double *input, double *output)
{
    Matvec matvec;
    
    //double *Ax   = new double[m_p];
    
    matvec.matvec_A_forward(n_p, m_p, G, s_val_p, input, output);
    
    matvec.matvec_A_transpose(n_p, m_p, G, s_val_p, output, output);
    
    for (long long i = 0; i <n_p; i++)
        output[i] = (hessian_pHuber[i])*input[i] + output[i];
    
    //delete[] Ax;
}

int PdNCG::sgn(double val)
{
    if (val < 0)
        return -1;
    else if (val > 0)
        return 1;
    else
        return 0;
}

double PdNCG::get_wall_time()
{
    struct timeval time;
    if (gettimeofday(&time,NULL))
    {
        //  Handle error
        return 0;
    }
    return (double)time.tv_sec + (double)time.tv_usec * .000001;
}

double PdNCG::get_cpu_time()
{
    return (double)clock() / CLOCKS_PER_SEC;
}